"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.assertInstructionNode = exports.isInstructionNode = exports.InstructionNode = void 0;
const utils_1 = require("../utils");
const TypeNode_1 = require("./TypeNode");
const TypeStructNode_1 = require("./TypeStructNode");
const TypeStructFieldNode_1 = require("./TypeStructFieldNode");
const ValueNode_1 = require("./ValueNode");
const TypeDefinedLinkNode_1 = require("./TypeDefinedLinkNode");
class InstructionNode {
    constructor(metadata, accounts, args, extraArgs, subInstructions) {
        this.nodeClass = 'InstructionNode';
        const bytes = metadata.bytesCreatedOnChain;
        this.metadata = {
            ...metadata,
            name: (0, utils_1.mainCase)(metadata.name),
            bytesCreatedOnChain: bytes && 'name' in bytes
                ? { ...bytes, name: (0, utils_1.mainCase)(bytes.name) }
                : bytes,
            argDefaults: Object.fromEntries(Object.entries(metadata.argDefaults).map(([key, value]) => {
                const newValue = { ...value };
                if ('name' in newValue) {
                    newValue.name = (0, utils_1.mainCase)(newValue.name);
                }
                if (newValue.kind === 'resolver') {
                    newValue.dependsOn = newValue.dependsOn.map((dep) => ({
                        ...dep,
                        name: (0, utils_1.mainCase)(dep.name),
                    }));
                }
                return [(0, utils_1.mainCase)(key), newValue];
            })),
        };
        this.accounts = accounts.map((account) => {
            const { defaultsTo } = account;
            if ((defaultsTo === null || defaultsTo === void 0 ? void 0 : defaultsTo.kind) === 'account') {
                defaultsTo.name = (0, utils_1.mainCase)(defaultsTo.name);
            }
            else if ((defaultsTo === null || defaultsTo === void 0 ? void 0 : defaultsTo.kind) === 'program') {
                defaultsTo.program.name = (0, utils_1.mainCase)(defaultsTo.program.name);
            }
            else if ((defaultsTo === null || defaultsTo === void 0 ? void 0 : defaultsTo.kind) === 'pda') {
                defaultsTo.pdaAccount = (0, utils_1.mainCase)(defaultsTo.pdaAccount);
                defaultsTo.seeds = Object.fromEntries(Object.entries(defaultsTo.seeds).map(([key, seed]) => [
                    (0, utils_1.mainCase)(key),
                    seed.kind === 'value'
                        ? seed
                        : { ...seed, name: (0, utils_1.mainCase)(seed.name) },
                ]));
            }
            else if ((defaultsTo === null || defaultsTo === void 0 ? void 0 : defaultsTo.kind) === 'resolver') {
                defaultsTo.name = (0, utils_1.mainCase)(defaultsTo.name);
                defaultsTo.dependsOn = defaultsTo.dependsOn.map((dep) => ({
                    ...dep,
                    name: (0, utils_1.mainCase)(dep.name),
                }));
            }
            return { ...account, name: (0, utils_1.mainCase)(account.name), defaultsTo };
        });
        this.args = args;
        this.extraArgs = extraArgs;
        this.subInstructions = subInstructions;
    }
    static fromIdl(idl) {
        var _a, _b, _c, _d, _e;
        const idlName = (_a = idl.name) !== null && _a !== void 0 ? _a : '';
        const name = (0, utils_1.camelCase)(idlName);
        const useProgramIdForOptionalAccounts = (_b = idl.defaultOptionalAccounts) !== null && _b !== void 0 ? _b : false;
        const metadata = {
            name,
            idlName,
            docs: (_c = idl.docs) !== null && _c !== void 0 ? _c : [],
            internal: false,
            bytesCreatedOnChain: null,
            argDefaults: {},
        };
        const accounts = ((_d = idl.accounts) !== null && _d !== void 0 ? _d : []).map((account) => {
            var _a, _b, _c, _d, _e, _f;
            const isOptional = (_b = (_a = account.optional) !== null && _a !== void 0 ? _a : account.isOptional) !== null && _b !== void 0 ? _b : false;
            return {
                name: (0, utils_1.camelCase)((_c = account.name) !== null && _c !== void 0 ? _c : ''),
                isWritable: (_d = account.isMut) !== null && _d !== void 0 ? _d : false,
                isSigner: account.isOptionalSigner
                    ? 'either'
                    : (_e = account.isSigner) !== null && _e !== void 0 ? _e : false,
                isOptional,
                description: (_f = account.desc) !== null && _f !== void 0 ? _f : '',
                defaultsTo: isOptional && useProgramIdForOptionalAccounts
                    ? { kind: 'programId' }
                    : null,
            };
        });
        let args = TypeStructNode_1.TypeStructNode.fromIdl({
            kind: 'struct',
            name: name ? `${name}InstructionData` : '',
            fields: (_e = idl.args) !== null && _e !== void 0 ? _e : [],
        });
        if (idl.discriminant) {
            const discriminatorField = new TypeStructFieldNode_1.TypeStructFieldNode({
                name: 'discriminator',
                docs: [],
                defaultsTo: {
                    strategy: 'omitted',
                    value: (0, ValueNode_1.vScalar)(idl.discriminant.value),
                },
            }, (0, TypeNode_1.createTypeNodeFromIdl)(idl.discriminant.type));
            args = new TypeStructNode_1.TypeStructNode(args.name, [
                discriminatorField,
                ...args.fields,
            ]);
        }
        return new InstructionNode(metadata, accounts, args, null, []);
    }
    accept(visitor) {
        return visitor.visitInstruction(this);
    }
    getAllSubInstructions() {
        return this.subInstructions.flatMap((subInstruction) => [
            subInstruction,
            ...subInstruction.getAllSubInstructions(),
        ]);
    }
    get name() {
        return this.metadata.name;
    }
    get docs() {
        return this.metadata.docs;
    }
    get hasLinkedArgs() {
        return (0, TypeDefinedLinkNode_1.isTypeDefinedLinkNode)(this.args);
    }
    get hasLinkedExtraArgs() {
        return (0, TypeDefinedLinkNode_1.isTypeDefinedLinkNode)(this.extraArgs);
    }
    get hasAccounts() {
        return this.accounts.length > 0;
    }
    get hasData() {
        if ((0, TypeDefinedLinkNode_1.isTypeDefinedLinkNode)(this.args))
            return true;
        return this.args.fields.length > 0;
    }
    get hasArgs() {
        if ((0, TypeDefinedLinkNode_1.isTypeDefinedLinkNode)(this.args))
            return true;
        const nonOmittedFields = this.args.fields.filter((field) => { var _a; return ((_a = field.metadata.defaultsTo) === null || _a === void 0 ? void 0 : _a.strategy) !== 'omitted'; });
        return nonOmittedFields.length > 0;
    }
    get hasExtraArgs() {
        var _a, _b;
        if ((0, TypeDefinedLinkNode_1.isTypeDefinedLinkNode)(this.extraArgs))
            return true;
        const nonOmittedFields = (_b = (_a = this.extraArgs) === null || _a === void 0 ? void 0 : _a.fields.filter((field) => { var _a; return ((_a = field.metadata.defaultsTo) === null || _a === void 0 ? void 0 : _a.strategy) !== 'omitted'; })) !== null && _b !== void 0 ? _b : [];
        return nonOmittedFields.length > 0;
    }
    get hasAnyArgs() {
        return this.hasArgs || this.hasExtraArgs;
    }
    get hasArgDefaults() {
        return Object.keys(this.metadata.argDefaults).length > 0;
    }
    get hasArgResolvers() {
        return Object.values(this.metadata.argDefaults).some(({ kind }) => kind === 'resolver');
    }
    get hasAccountResolvers() {
        return this.accounts.some(({ defaultsTo }) => (defaultsTo === null || defaultsTo === void 0 ? void 0 : defaultsTo.kind) === 'resolver');
    }
    get hasByteResolver() {
        var _a;
        return ((_a = this.metadata.bytesCreatedOnChain) === null || _a === void 0 ? void 0 : _a.kind) === 'resolver';
    }
    get hasResolvers() {
        return (this.hasArgResolvers || this.hasAccountResolvers || this.hasByteResolver);
    }
}
exports.InstructionNode = InstructionNode;
function isInstructionNode(node) {
    return !!node && node.nodeClass === 'InstructionNode';
}
exports.isInstructionNode = isInstructionNode;
function assertInstructionNode(node) {
    var _a;
    if (!isInstructionNode(node)) {
        throw new Error(`Expected InstructionNode, got ${(_a = node === null || node === void 0 ? void 0 : node.nodeClass) !== null && _a !== void 0 ? _a : 'null'}.`);
    }
}
exports.assertInstructionNode = assertInstructionNode;
//# sourceMappingURL=InstructionNode.js.map